#!/bin/sh

NO_REBOOT="0"
CHECK_VERSION="0"

LOCK_FILE=/tmp/lock_multiple_upgrade
BOOT_VERSION_FILE=/tmp/boot_version_file
ENV_SIZE=0x80000
CUR_BOOTBANK=
CUR_BROKEN_BANK_1=
CUR_BROKEN_BANK_2=
CUR_STAGE=0
BOOTLOADER_IS_FOUND=0
FIRMWARE_IS_FOUND=0
BOOTLOADER_INSTALL_REQUIRED=0
FIRMWARE_INSTALL_REQUIRED=0

MTD_START_IDX=0
MTD_END_IDX=11

fip_file="fip"
env_file="u-boot-env"
dtb_file="dt-blob"
kernel_file="kernel"
rootfs_file="root"

firmware_mtd_list="$dtb_file $kernel_file $rootfs_file"
bootloader_mtd_list="$fip_file $env_file"

function read_env_variables {
  echo "Reading variables..." > /dev/console

  CUR_BOOTBANK=$(($(ubootenv -g "cur_bootbank" | cut -d '=' -f2)))
  CUR_BROKEN_BANK_1=$(($(ubootenv -g "broken_bank_1" | cut -d '=' -f2)))
  CUR_BROKEN_BANK_2=$(($(ubootenv -g "broken_bank_2" | cut -d '=' -f2)))
  
  echo "  cur_bootbank is ${CUR_BOOTBANK}" > /dev/console
  echo "  broken_bank_1 is ${CUR_BROKEN_BANK_1}" > /dev/console
  echo "  broken_bank_2 is ${CUR_BROKEN_BANK_2}" > /dev/console
}

function save_env_variables {
  echo "Saving variables..." > /dev/console

  ubootenv -s cur_bootbank ${CUR_BOOTBANK}
  ubootenv -s broken_bank_1 ${CUR_BROKEN_BANK_1}
  ubootenv -s broken_bank_2 ${CUR_BROKEN_BANK_2}

  if [ ! "$(ubootenv -g mtd_bootargs | grep 'rootflags=sync')" ]
  then
    ubootenv -s mtd_bootargs "\${mtdparts} ubi.mtd=\${rootpartidx} root=ubi0:rootfs rootfstype=ubifs rootflags=sync rw"
  fi
}

function check_md5_match {
  file_base=$1

  md5sum1=$(awk '{ print $1 }' < "./$file_base.md5sum")
  md5sum2=$(md5sum "./bin.$file_base" | awk '{ print $1 }')

  if [ "$md5sum1" != "$md5sum2" ]; then
    echo "[Error] md5sum's isn't equal!" > /dev/console
    exit 1
  fi
}

# $1 arg == mtd
# $2 arg == action (0 == write, 1 == erase)
function upgrade_failed {
  CUR_STAGE=$(($CUR_STAGE + 1))
  MTD=$1
  ACTION=$2

  if [[ "$ACTION" == "0" ]]
  then
    echo "  Stage ${CUR_STAGE}: writing ${MTD} has been FAILED" > /dev/console
  else
    echo "  Stage ${CUR_STAGE}: erasing ${MTD} has been FAILED" > /dev/console
  fi

  echo "[Error] Firmware upgrade has been FAILED. Please check files or permissions!" > /dev/console
  exit 1
}

# $1 arg == mtd
# $2 action (0 == write, 1 == erase)
function log_upgrade_ok {
  CUR_STAGE=$(($CUR_STAGE + 1))
  MTD=$1
  ACTION=$2

  if [[ "$ACTION" == "0" ]]
  then
    echo "  Stage ${CUR_STAGE}: writing ${MTD} has been successfully completed" > /dev/console
  else
    echo "  Stage ${CUR_STAGE}: erasing ${MTD} has been successfully completed" > /dev/console
  fi
}

# $1 arg == mtd
# $2 arg == BLOCK_COUNT
function erase_flash {
  MTD=$1
  BLOCK_COUNT=$2

  flash_erase "${MTD}" 0 "${BLOCK_COUNT}" > /dev/null

  if [ "$?" -ne "0" ]; then
    upgrade_failed ${MTD} 1
  else
    log_upgrade_ok ${MTD} 1
  fi
}

check_files_exist() {
  for i in $bootloader_mtd_list; do
    if [ -f "./bin.$i" ]; then
      if [ ! -f "./$i.md5sum" ]; then
        echo "[Error] ./$i.md5sum doesn't exist!" > /dev/console
        exit 1
      fi
      check_md5_match "$i"

      if [ "$i" == "$fip_file" ] || [ "$i" == "$env_file" ]; then
        BOOTLOADER_IS_FOUND=1
        BOOTLOADER_INSTALL_REQUIRED=1
      fi
    fi
  done

  for i in $firmware_mtd_list; do
    if [ -f "./bin.$i" ]; then
      if [ ! -f "./$i.md5sum" ]; then
        echo "[Error] ./$i.md5sum doesn't exist!" > /dev/console
        exit 1
      fi
      check_md5_match "$i"

      if [ "$i" == "$dtb_file" ] || [ "$i" == "$kernel_file" ] || [ "$i" == "$rootfs_file" ]; then
        FIRMWARE_IS_FOUND=1
        FIRMWARE_INSTALL_REQUIRED=1
      fi
    fi
  done

  if [ ! -f "./version" ]; then
    echo "[Error] ./version doesn't exist!" > /dev/console
    exit 1
  fi

  if [ "$FIRMWARE_IS_FOUND" -eq "0" ] && [ "$BOOTLOADER_IS_FOUND" -eq "0" ]; then
    echo "[Error] archive content isn't correct!" > /dev/console
    exit 1
  fi
}

check_actual_version() {
    if [ "$FIRMWARE_IS_FOUND" -eq "1" ] ; then
      cmp -s version /etc/version

      if [ "$?" == 0 ]; then
        echo "The actual firmware version is already installed" > /dev/console
        FIRMWARE_INSTALL_REQUIRED=0
      fi
    fi

    if [ "$BOOTLOADER_IS_FOUND" -eq "1" ] ; then
      ubootenv -g ver | cut -d '=' -f 2 > $BOOT_VERSION_FILE
      cmp -s version $BOOT_VERSION_FILE

      if [ "$?" == 0 ]; then
        echo "The actual bootloader version is already installed" > /dev/console
        BOOTLOADER_INSTALL_REQUIRED=0
      fi

      rm $BOOT_VERSION_FILE
    fi
}

handle_version() {
	if [[ $CHECK_VERSION -ne 0 ]]; then
	  echo "Check actual firmware version" > /dev/console
	  check_actual_version
	fi
}

check_upgrade_needing() {
  handle_version

  if [ "$FIRMWARE_IS_FOUND" -eq "1" ] && [ "$BOOTLOADER_IS_FOUND" -eq "0" ]; then
    if [ "$FIRMWARE_INSTALL_REQUIRED" -eq "0" ]; then
      exit 2
    fi
  fi

  if [ "$FIRMWARE_IS_FOUND" -eq "0" ] && [ "$BOOTLOADER_IS_FOUND" -eq "1" ]; then
    if [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "0" ]; then
      exit 2
    fi
  fi

  if [ "$FIRMWARE_IS_FOUND" -eq "1" ] && [ "$BOOTLOADER_IS_FOUND" -eq "1" ]; then
    if [ "$FIRMWARE_INSTALL_REQUIRED" -eq "0" ] && [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "0" ]; then
      exit 2
    fi
  fi

  VERSION="$(cat ./version)"
  echo "Version $VERSION will be installed" > /dev/console
}


# $1 arg == mtd
# $2 arg == path to file
function write_flash {
  MTD=$1
  WRITE_PATH=$2

  nandwrite -pk "${MTD}" "${WRITE_PATH}" > /dev/null

  if [ "$?" -ne "0" ]; then
    upgrade_failed ${MTD} 0
  else
    log_upgrade_ok ${MTD} 0
  fi
}

function upgrade_firmware {
  x=0

  for i in $firmware_mtd_list; do
    if [ -f "./bin.$i" ]; then
      echo "./bin.$i" > /dev/console

      cur_mtd_idx=$(($x + 10 - 3 * $CUR_BOOTBANK))

      MEM=$(($(cat /proc/mtd | grep "mtd${cur_mtd_idx}" | awk '{pre = "0x"; print pre $2 }')))
      BLOCK_SIZE=$(($(cat /proc/mtd | grep "mtd${cur_mtd_idx}" | awk '{pre = "0x"; print pre $3 }')))
      BLOCK_COUNT=$(expr $MEM / $BLOCK_SIZE)

      erase_flash "/dev/mtd${cur_mtd_idx}" $BLOCK_COUNT
      write_flash "/dev/mtd${cur_mtd_idx}" "./bin.$i"
    fi
    x=$((x+1))
  done
}

function upgrade_bootloader {

  for i in $bootloader_mtd_list; do
    if [ -f "./bin.$i" ]; then

      echo "./bin.$i" > /dev/console

      MTD_IDX=$(cat /proc/mtd | grep "\"${i}\"" | awk '{print $1}' | grep -o '[0-9]\+')

      if [ -z "${MTD_IDX}" ]; then

        echo "mtd index not found, $i";

      else
       
        if (( $MTD_IDX >= $MTD_START_IDX &&  $MTD_IDX < $MTD_END_IDX )); then 

          MEM=$(($(cat /proc/mtd | grep "mtd${MTD_IDX}:" | awk '{pre = "0x"; print pre $2 }')))
          BLOCK_SIZE=$(($(cat /proc/mtd | grep "mtd${MTD_IDX}:" | awk '{pre = "0x"; print pre $3 }')))
          BLOCK_COUNT=$(expr $MEM / $BLOCK_SIZE)

          erase_flash "/dev/mtd${MTD_IDX}" $BLOCK_COUNT

          if [ "$i" == "$env_file" ]; then
            nandwrite -s 0 -pk /dev/mtd${MTD_IDX} "./bin.$i" > /dev/null
            nandwrite -s ${ENV_SIZE} -pk /dev/mtd${MTD_IDX} "./bin.$i" > /dev/null
            log_upgrade_ok "/dev/mtd${MTD_IDX}" 0
          else
            write_flash "/dev/mtd${MTD_IDX}" "./bin.$i"
          fi

        else

          echo "bad mtd index, skip upgrade bin.$i"

        fi
      fi
    fi
  done
}

function upgrade_process {
  #Step 1. Checking
  check_files_exist
  check_upgrade_needing

  #Step 2. Send upgrade notification
  if [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "1" ] || [ "$FIRMWARE_INSTALL_REQUIRED" -eq "1" ]; then
    setupgradestatus 1
  fi

  #Step 3. Start the upgrade process
  if [ "$BOOTLOADER_INSTALL_REQUIRED" -eq "1" ]; then
    echo "===== Start of BOOTLOADER upgrade =====" > /dev/console

    read_env_variables

    upgrade_bootloader

    echo "Recovering variables" > /dev/console

    save_env_variables

    echo "===== End of BOOTLOADER upgrade =====" > /dev/console
  fi

  if [ "$FIRMWARE_INSTALL_REQUIRED" -eq "1" ]; then
    echo "===== Start of FIRMWARE upgrade =====" > /dev/console

    read_env_variables

    upgrade_firmware

    # We need to update some uboot environments here
    if [ "$CUR_BOOTBANK" -eq "1" ]; then
      CUR_BOOTBANK=2
    else
      CUR_BOOTBANK=1
    fi

    if [ "$CUR_BOOTBANK" -eq "1" ]; then
      CUR_BROKEN_BANK_1=0
    else
      CUR_BROKEN_BANK_2=0
    fi

    save_env_variables

    echo "===== End of FIRMWARE upgrade =====" > /dev/console
  fi

  echo "Upgrade has been successfully completed. Well done!" > /dev/console
}

###############################################################################################

if [ -f ${LOCK_FILE} ]; then
  echo "[Warning] You need to reboot to complete the upgrade process!" > /dev/console
  exit 1
fi

while getopts nv flag
do
    case "${flag}" in
        n) NO_REBOOT="1";;
        v) CHECK_VERSION="1";;
    esac
done

upgrade_process

echo "1" > ${LOCK_FILE}

if [[ $NO_REBOOT -ne 1 ]]; then
  echo "The system will be rebooted" > /dev/console
  reboot
fi
